<?php
/*******************************************************************************
hmm_ShowRanges - derived from nwsw_PitchMod.php Version 1.0

Inspired by nwsw_PitchMod and by adp_Ranges, and made necessary by my plan
to start arranging for concert band, this scripts highlights notes that 
are outside ranges defined by an initial "description chord".

Copyright  2014 by Harald M. Mller
All Rights Reserved

HISTORY:
================================================================================
[2014-10-27] Version 1.0:  Initial release
*******************************************************************************/
require_once("lib/nwc2clips.inc");

define("SHOW_RANGE_VERSION","1.0");

if (array_key_exists(1,$argv) && strstr('help',strtolower($argv[1]))) doHelp();

doReport("\nProcessing input\n");

$clip = new NWC2Clip('php://stdin');
echo $clip->GetClipHeader()."\n";

$PlayContext = new NWC2PlayContext();

$replacements = array(); // of NWC2ClipItem

foreach ($clip->Items as $item) {
	$o = new NWC2ClipItem($item,true);

	if ($o->IsContextInfo()) {
		$PlayContext->UpdateContext($o);
		continue;
		}

  if (empty($replacements)) if (SetRangeLimitsIfRangeChord()) {
		$PlayContext->UpdateContext($o);
		echo $item;
		continue;
    }

	$isChanged = false;

	if (in_array($o->GetObjType(),array('Note','Chord','RestChord'))) {
		$o_new = new NWC2ClipItem($item,true);

		for ($loop=0;$loop<2;$loop++) {
			$Opt_inc = $loop ? "2" : "";
			$OptName_Pos = "Pos".$Opt_inc;
			$OptName_Dur = "Dur".$Opt_inc;

			$pitchpos = &$o_new->GetTaggedOpt($OptName_Pos);
			if (is_string($pitchpos)) ProcessNote($pitchpos);
			else if (is_array($pitchpos)) foreach ($pitchpos as &$notepitchTxt) ProcessNote($notepitchTxt);
			}

		if ($isChanged) echo $o_new->ReconstructClipText()."\n";
		}

	if (!$isChanged) echo $item;

	$PlayContext->UpdateContext($o);
	}

echo NWC2_ENDCLIP."\n";
//exit(NWC2RC_REPORT);
//exit(NWC2RC_ERROR);
exit(NWC2RC_SUCCESS);

function SetRangeLimitsIfRangeChord()
{
  global $o,$replacements,$PlayContext,$max,$minMidi;
  
  if ($o->GetObjType()=="Chord") {    
    $chordnotes = array_values($o->GetTaggedOptAsArray("Pos"));
    $pos2Chordnotes = $o->GetTaggedOptAsArray("Pos2");
    if ($pos2Chordnotes) {
      $chordnotes = array_merge($chordnotes, array_values($pos2Chordnotes));
      }

    $createNWC2NotePitchPos = function($t) { return new NWC2NotePitchPos($t); };
    $notepitchObjs = array_map($createNWC2NotePitchPos, $chordnotes);
    $compare = function($na,$nb) { global $PlayContext; return $PlayContext->GetNoteMidiPitch($nb) - $PlayContext->GetNoteMidiPitch($na); };

    // hold NWC2NotePitchPos objects ordered by midi value (maximal midi value at [0])
    usort($notepitchObjs, $compare);

    ////var_dump(">>6", $notepitchObjs);
    
    $dump = "";
    while ($max = array_shift($notepitchObjs)) {
      // find opposite NWC2NotePitchPos
      $maxMidi = $PlayContext->GetNoteMidiPitch($max);
      $minIndex = count($notepitchObjs) - 1;
      if ($minIndex < 0) break;
      else {
        $minMidi = $PlayContext->GetNoteMidiPitch($notepitchObjs[$minIndex]);
        $matchesMaxAndMinMidi = function($n) { 
        	global $max,$minMidi,$PlayContext; 
        	////var_dump(">>1",$n,$max,$n->Notehead,$max->Notehead,"<<1");
        	return $PlayContext->GetNoteMidiPitch($n) == $minMidi && $n->Notehead == $max->Notehead && $n->Color == $max->Color; };
        
        // If we find a matching min, ...
        $matchingMins = array_values(array_filter($notepitchObjs, $matchesMaxAndMinMidi));
        
        if (count($matchingMins) > 0) {          
          $matchingMin = $matchingMins[0];
          ////var_dump(">>3",$matchingMins,$matchingMin);
          unset($notepitchObjs[$minIndex]);
        
          // ... we record $max for all midi notes above max and below min that are not yet assigned.
          for ($i = $maxMidi+1; $i < 128; $i++) {
            if (array_key_exists($i, $replacements)) {
              break;
              }
            $replacements[$i] = $max;
            $dump .= "+".(1000+$i);
            }
          for ($i = $minMidi-1; $i >= 0; $i--) {
            if (array_key_exists($i, $replacements)) {
              break;
              }
            $replacements[$i] = $max;
            $dump .= "-".(1000+$i);
            }
            $dump .= "|";
          }
        else {
          // no match found: This is not a valid range chord!
          // Should we output something here? I hope I can do without ...
          ////var_dump(">>7", $replacements);

          $replacements = array();
          return false;
          }
        }
      } // while
    
    // We did it!
    return true;
    }
  else {
    // Not a chord.
    return false;
    }
}

function ProcessNote(&$notepitchTxt)
{
  global $PlayContext,$o,$OptName_Pos,$OptName_Dur,$isChanged,$replacements;

  if ($o->GetTaggedOpt("Visibility") == "Never") return;

	static $accmap = array("v"=>"bb","b"=>"b","n"=>"n","#"=>"#","x"=>"x");

	$notepitchObj = new NWC2NotePitchPos($notepitchTxt);
	$notemidipitch = $PlayContext->GetNoteMidiPitch($notepitchObj);

  // Modify if necessary
  if (array_key_exists($notemidipitch,$replacements)) {
    $nh = $replacements[$notemidipitch]->Notehead;
    $co = $replacements[$notemidipitch]->Color;
    if ($notepitchObj->Notehead != $nh || $notepitchObj->Color != $co) {
      $notepitchObj->Notehead = $nh;
      $notepitchObj->Color = $co;
    	$notepitchTxt = $notepitchObj->ReconstructClipText();	
	    $isChanged = true;    
      }
    }
}


function doHelp($s="")
{
	if ($s) echo $s."\n";

	echo "Show range, Version ".SHOW_RANGE_VERSION."\n\n";

	echo <<<__EOHELPTEXT
Usage: hmm_ShowRanges.php

This script expects that the first chord of the input is a 'marker chord' that defines ranges with identical noteheads and/or colors 'inside out'.
Notes outside a range are then marked accordingly, i.e. their notehead and/or color is changed.

__EOHELPTEXT;

	exit(NWC2RC_REPORT);
}

function doReport($s)
{
	if (defined("DO_GEN_REPORT"))	fputs(STDERR, $s);
}

?>